from contextlib import contextmanager
from dataclasses import dataclass
from importlib import resources
from pathlib import Path
from prosperity2bt.datamodel import Symbol, Trade
from typing import Iterator, Optional

LIMITS = {
    "AMETHYSTS": 20,
    "STARFRUIT": 20,
}

@dataclass
class PriceRow:
    day: int
    timestamp: int
    product: Symbol
    bid_prices: list[int]
    bid_volumes: list[int]
    ask_prices: list[int]
    ask_volumes: list[int]
    mid_price: float
    profit_loss: float

@dataclass
class DayData:
    round: int
    day: int

    prices: list[PriceRow]
    trades: list[Trade]

@contextmanager
def wrap_in_context_manager(path: Path) -> Iterator[Path]:
    yield path

def get_column_values(columns: list[str], indices: list[int]) -> list[int]:
    values = []

    for index in indices:
        value = columns[index]
        if value == "":
            break

        values.append(int(value))

    return values

def read_day_data(data_root: Optional[Path], round: int, day: int) -> DayData:
    prices_file_name = f"prices_round_{round}_day_{day}.csv"
    trades_file_name = f"trades_round_{round}_day_{day}_nn.csv"

    if data_root is not None:
        round_dir = data_root / f"round{round}"
        prices_file = wrap_in_context_manager(round_dir / prices_file_name)
        trades_file = wrap_in_context_manager(round_dir / trades_file_name)
    else:
        round_dir = resources.files(f"prosperity2bt.resources.round{round}")
        prices_file = resources.as_file(round_dir / prices_file_name)
        trades_file = resources.as_file(round_dir / trades_file_name)

    prices = []
    with prices_file as file:
        for line in file.read_text(encoding="utf-8").splitlines()[1:]:
            columns = line.split(";")

            prices.append(PriceRow(
                day=int(columns[0]),
                timestamp=int(columns[1]),
                product=columns[2],
                bid_prices=get_column_values(columns, [3, 5, 7]),
                bid_volumes=get_column_values(columns, [4, 6, 8]),
                ask_prices=get_column_values(columns, [9, 11, 13]),
                ask_volumes=get_column_values(columns, [10, 12, 14]),
                mid_price=float(columns[15]),
                profit_loss=float(columns[16]),
            ))

    trades = []
    with trades_file as file:
        for line in file.read_text(encoding="utf-8").splitlines()[1:]:
            columns = line.split(";")

            trades.append(Trade(
                symbol=columns[3],
                price=int(float(columns[5])),
                quantity=int(columns[6]),
                buyer=columns[1],
                seller=columns[2],
                timestamp=int(columns[0]),
            ))

    return DayData(round, day, prices, trades)

def read_round_data(data_root: Optional[Path], round: int) -> list[DayData]:
    if data_root is not None:
        files = (data_root / f"round{round}").iterdir()
    else:
        files = resources.files(f"prosperity2bt.resources.round{round}").iterdir()

    days = []
    for file in files:
        if file.name.startswith("prices_round_"):
            day = int(file.stem.split("_")[-1])
            days.append(read_day_data(data_root, round, day))

    return days



import json
import sys
import webbrowser
from argparse import ArgumentParser
from collections import defaultdict
from datetime import datetime
from functools import partial, reduce
from http.server import HTTPServer, SimpleHTTPRequestHandler
from importlib import import_module, metadata
from pathlib import Path
from prosperity2bt.core import DayResult, run_backtest
from prosperity2bt.data import DayData, read_day_data, read_round_data
from prosperity2bt.datamodel import ProsperityEncoder
from typing import Any, Optional

def parse_algorithm(algorithm: str) -> tuple[Any, Any]:
    algorithm_path = Path(algorithm).expanduser().resolve()

    sys.path.append(str(algorithm_path.parent))
    return import_module(algorithm_path.stem).Trader

def parse_days(days: list[str], data_root: Optional[str]) -> list[DayData]:
    parsed_days = []

    if data_root is not None:
        data_root = Path(data_root).expanduser().resolve()

    for arg in days:
        if "-" in arg:
            round, day = map(int, arg.split("-", 1))
            parsed_days.append(read_day_data(data_root, round, day))
        else:
            round = int(arg)
            parsed_days.extend(read_round_data(data_root, round))

    return parsed_days

def parse_out(out: Optional[str]) -> Path:
    if out is not None:
        return Path(out).expanduser().resolve()
    else:
        timestamp = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
        return Path.cwd() / "backtests" / f"{timestamp}.log"

def offset_sandbox_log_row(row: Any, timestamp_offset: int) -> Any:
    old_timestamp = row["timestamp"]
    new_timestamp = old_timestamp + timestamp_offset

    row["lambdaLog"] = row["lambdaLog"].replace(f"[[{old_timestamp},", f"[[{new_timestamp},")
    row["timestamp"] += timestamp_offset

    return row

def offset_trade(trade: Any, timestamp_offset: int) -> Any:
    trade["timestamp"] += timestamp_offset
    return trade

def merge_results(a: DayResult, b: DayResult, merge_profit_loss: bool) -> DayResult:
    sandbox_logs = a.sandbox_logs[:]
    activity_logs = a.activity_logs[:]
    trades = a.trades[:]

    a_last_timestamp = a.activity_logs[-1].timestamp
    timestamp_offset = a_last_timestamp + 100

    sandbox_logs.extend([offset_sandbox_log_row(row, timestamp_offset) for row in b.sandbox_logs])

    if merge_profit_loss:
        profit_loss_offsets = defaultdict(float)
        for row in reversed(a.activity_logs):
            if row.timestamp != a_last_timestamp:
                break

            profit_loss_offsets[row.columns[2]] = row.columns[-1]

        activity_logs.extend([
            row.with_offset(timestamp_offset, profit_loss_offsets[row.columns[2]])
            for row in b.activity_logs
        ])
    else:
        activity_logs.extend([row.with_offset(timestamp_offset, 0) for row in b.activity_logs])

    trades.extend([offset_trade(trade, timestamp_offset) for trade in b.trades])

    return DayResult(a.round, a.day, sandbox_logs, activity_logs, trades)

def write_output(output_file: Path, merged_results: DayResult) -> None:
    output_file.parent.mkdir(parents=True, exist_ok=True)
    with output_file.open("w+", encoding="utf-8") as file:
        file.write("Sandbox logs:\n")
        file.write("\n".join(json.dumps(row, indent=2) for row in merged_results.sandbox_logs))

        file.write("\n\n\n\nActivities log:\n")
        file.write("day;timestamp;product;bid_price_1;bid_volume_1;bid_price_2;bid_volume_2;bid_price_3;bid_volume_3;ask_price_1;ask_volume_1;ask_price_2;ask_volume_2;ask_price_3;ask_volume_3;mid_price;profit_and_loss\n")
        file.write("\n".join(map(str, merged_results.activity_logs)))

        file.write("\n\n\n\n\nTrade History:\n")
        file.write(json.dumps(merged_results.trades, cls=ProsperityEncoder, indent=2))

def print_overall_summary(results: list[DayResult]) -> None:
    print(f"Profit summary:")

    total_profit = 0
    for result in results:
        last_timestamp = result.activity_logs[-1].timestamp

        profit = 0
        for row in reversed(result.activity_logs):
            if row.timestamp != last_timestamp:
                break

            profit += row.columns[-1]

        print(f"Round {result.round} day {result.day}: {profit:,.0f}")
        total_profit += profit

    print(f"Total profit: {total_profit:,.0f}\n")

class HTTPRequestHandler(SimpleHTTPRequestHandler):
    def end_headers(self) -> None:
        self.send_header("Access-Control-Allow-Origin", "*")
        return super().end_headers()

    def log_message(self, format: str, *args: Any) -> None:
        return

def open_visualizer(output_file: Path) -> None:
    http_handler = partial(HTTPRequestHandler, directory=output_file.parent)
    http_server = HTTPServer(("localhost", 0), http_handler)

    webbrowser.open(f"https://jmerle.github.io/imc-prosperity-2-visualizer/?open=http://localhost:{http_server.server_port}/{output_file.name}")
    http_server.handle_request()
    http_server.handle_request()

def format_path(path: Path) -> str:
    cwd = Path.cwd()
    if path.is_relative_to(cwd):
        return str(path.relative_to(cwd))
    else:
        return str(path)

def main() -> None:
    parser = ArgumentParser(prog="prosperity2bt", description="Run a backtest.")
    parser.add_argument("algorithm", type=str, help="path to the Python file containing the algoritm to backtest")
    parser.add_argument("days", type=str, nargs="+", help="the days to backtest on (<round>-<day> for a single day, <round> for all days in a round)")
    parser.add_argument("--merge-pnl", action="store_true", help="merge profit and loss across days")
    parser.add_argument("--vis", action="store_true", help="open backtest result in visualizer when done")
    parser.add_argument("--out", type=str, help="path to save output log to (defaults to backtests/<timestamp>.log)")
    parser.add_argument("--data", type=str, help="path to data directory (must look similar in structure to https://github.com/jmerle/imc-prosperity-2-backtester/tree/master/prosperity2bt/resources)")
    parser.add_argument("--print", action="store_true", help="print the trader's output to stdout while it's running")
    parser.add_argument("-v", "--version", action="version", version=f"%(prog)s {metadata.version(__package__)}")

    args = parser.parse_args()

    Trader = parse_algorithm(args.algorithm)
    days = parse_days(args.days, args.data)
    output_file = parse_out(args.out)

    results = [run_backtest(Trader(), day, args.print) for day in days]
    merged_results = reduce(lambda a, b: merge_results(a, b, args.merge_pnl), results)

    write_output(output_file, merged_results)

    if len(days) > 1:
        print_overall_summary(results)

    print(f"Successfully saved backtest results to {format_path(output_file)}")

    if args.vis:
        open_visualizer(output_file)

if __name__ == "__main__":
    main()



from collections import defaultdict
from contextlib import closing, redirect_stdout
from dataclasses import dataclass
from io import StringIO
from IPython.utils.io import Tee
from prosperity2bt.data import DayData, LIMITS, PriceRow
from prosperity2bt.datamodel import Observation, Order, OrderDepth, Symbol, Trade, TradingState
from typing import Any

@dataclass
class ActivityLogRow:
    columns: list[Any]

    @property
    def timestamp(self) -> int:
        return self.columns[1]

    def with_offset(self, timestamp_offset: int, profit_loss_offset: float) -> "ActivityLogRow":
        new_columns = self.columns[:]
        new_columns[1] += timestamp_offset
        new_columns[-1] += profit_loss_offset

        return ActivityLogRow(new_columns)

    def __str__(self) -> str:
        return ";".join(map(str, self.columns))

@dataclass
class DayResult:
    round: int
    day: int

    sandbox_logs: list[dict[str, Any]]
    activity_logs: list[ActivityLogRow]
    trades: list[Trade]

@dataclass
class MarketTrade:
    trade: Trade
    buy_quantity: int
    sell_quantity: int

def check_limits(
    tradable_products: list[str],
    orders_by_symbol: dict[Symbol, list[Order]],
    own_positions: dict[str, int],
) -> None:
    sandbox_log_lines = []
    for product in tradable_products:
        orders = orders_by_symbol.get(product, [])

        current_position = own_positions[product]

        total_long = sum(order.quantity for order in orders if order.quantity > 0)
        total_short = sum(abs(order.quantity) for order in orders if order.quantity < 0)

        if current_position + total_long > LIMITS[product] or current_position - total_short < -LIMITS[product]:
            sandbox_log_lines.append(f"Orders for product {product} exceeded limit of {LIMITS[product]} set")
            orders_by_symbol.pop(product)

    if len(sandbox_log_lines) == 0:
        return ""

    return "\n" + "\n".join(sandbox_log_lines)

def process_buy_order(
    timestamp: int,
    order: Order,
    order_depth: OrderDepth,
    own_positions: dict[Symbol, int],
    profit_loss_by_product: dict[Symbol, float],
    market_trades: list[MarketTrade],
) -> list[Trade]:
    new_trades = []

    price_matches = sorted((price for price in order_depth.sell_orders.keys() if price <= order.price), reverse=True)
    for price in price_matches:
        volume = min(order.quantity, abs(order_depth.sell_orders[price]))

        new_trades.append(Trade(order.symbol, order.price, volume, "SUBMISSION", "", timestamp))

        own_positions[order.symbol] += volume
        profit_loss_by_product[order.symbol] -= order.price * volume

        order_depth.sell_orders[price] += volume
        if order_depth.sell_orders[price] == 0:
            order_depth.sell_orders.pop(price)

        order.quantity -= volume
        if order.quantity == 0:
            return new_trades

    for trade in market_trades:
        if trade.sell_quantity == 0 or trade.trade.price > order.price:
            continue

        volume = min(order.quantity, trade.sell_quantity)

        new_trades.append(Trade(order.symbol, order.price, volume, "SUBMISSION", trade.trade.seller, timestamp))

        own_positions[order.symbol] += volume
        profit_loss_by_product[order.symbol] -= order.price * volume

        trade.sell_quantity -= volume

        order.quantity -= volume
        if order.quantity == 0:
            return new_trades

    return new_trades

def process_sell_order(
    timestamp: int,
    order: Order,
    order_depth: OrderDepth,
    own_positions: dict[Symbol, int],
    profit_loss_by_product: dict[Symbol, float],
    market_trades: list[MarketTrade],
) -> list[Trade]:
    new_trades = []

    price_matches = sorted(price for price in order_depth.buy_orders.keys() if price >= order.price)
    for price in price_matches:
        volume = min(abs(order.quantity), order_depth.buy_orders[price])

        new_trades.append(Trade(order.symbol, order.price, volume, "", "SUBMISSION", timestamp))

        own_positions[order.symbol] -= volume
        profit_loss_by_product[order.symbol] += order.price * volume

        order_depth.buy_orders[price] -= volume
        if order_depth.buy_orders[price] == 0:
            order_depth.buy_orders.pop(price)

        order.quantity += volume
        if order.quantity == 0:
            return new_trades

    for trade in market_trades:
        if trade.buy_quantity == 0 or trade.trade.price < order.price:
            continue

        volume = min(abs(order.quantity), trade.buy_quantity)

        new_trades.append(Trade(order.symbol, order.price, volume, trade.trade.buyer, "SUBMISSION", timestamp))

        own_positions[order.symbol] -= volume
        profit_loss_by_product[order.symbol] += order.price * volume

        trade.buy_quantity -= volume

        order.quantity += volume
        if order.quantity == 0:
            return new_trades

    return new_trades

def process_order(
    timestamp: int,
    order: Order,
    order_depths: dict[Symbol, OrderDepth],
    own_positions: dict[Symbol, int],
    profit_loss_by_product: dict[Symbol, float],
    market_trades: list[MarketTrade],
) -> list[Trade]:
    order_depth = order_depths[order.symbol]
    if order.quantity > 0:
        return process_buy_order(timestamp, order, order_depth, own_positions, profit_loss_by_product, market_trades)
    elif order.quantity < 0:
        return process_sell_order(timestamp, order, order_depth, own_positions, profit_loss_by_product, market_trades)
    else:
        return []

def trade_to_dict(trade: Trade) -> dict[str, Any]:
    return {
        "timestamp": trade.timestamp,
        "buyer": trade.buyer,
        "seller": trade.seller,
        "symbol": trade.symbol,
        "currency": "SEASHELLS",
        "price": trade.price,
        "quantity": trade.quantity,
    }

def create_activity_log_row(
    day: int,
    timestamp: int,
    product: str,
    price: PriceRow,
    profit_loss: float,
) -> ActivityLogRow:
    bid_prices_len = len(price.bid_prices)
    bid_volumes_len = len(price.bid_volumes)
    ask_prices_len = len(price.ask_prices)
    ask_volumes_len = len(price.ask_volumes)

    columns = [
        day,
        timestamp,
        product,
        price.bid_prices[0] if bid_prices_len > 0 else "",
        price.bid_volumes[0] if bid_volumes_len > 0 else "",
        price.bid_prices[1] if bid_prices_len > 1 else "",
        price.bid_volumes[1] if bid_volumes_len > 1 else "",
        price.bid_prices[2] if bid_prices_len > 2 else "",
        price.bid_volumes[2] if bid_volumes_len > 2 else "",
        price.ask_prices[0] if ask_prices_len > 0 else "",
        price.ask_volumes[0] if ask_volumes_len > 0 else "",
        price.ask_prices[1] if ask_prices_len > 1 else "",
        price.ask_volumes[1] if ask_volumes_len > 1 else "",
        price.ask_prices[2] if ask_prices_len > 2 else "",
        price.ask_volumes[2] if ask_volumes_len > 2 else "",
        price.mid_price,
        profit_loss,
    ]

    return ActivityLogRow(columns)

def print_backtest_summary(result: DayResult, tradable_products: list[str]) -> None:
    last_timestamp = result.activity_logs[-1].timestamp

    for product in sorted(tradable_products):
        product_profit = next(
            row.columns[-1] for row in reversed(result.activity_logs)
            if row.timestamp == last_timestamp and row.columns[2] == product
        )

        print(f"{product}: {product_profit:,.0f}")

    total_profit = 0
    for row in reversed(result.activity_logs):
        if row.timestamp != last_timestamp:
            break

        total_profit += row.columns[-1]

    print(f"Total profit: {total_profit:,.0f}\n")

def run_backtest(trader: Any, data: DayData, print_output: bool) -> DayResult:
    print(f"Backtesting {trader.__module__} on round {data.round} day {data.day}")

    result = DayResult(data.round, data.day, [], [], [])
    trader_data = ""

    prices_by_timestamp = defaultdict(dict)
    for row in data.prices:
        prices_by_timestamp[row.timestamp][row.product] = row

    trades_by_timestamp = defaultdict(lambda: defaultdict(list))
    for trade in data.trades:
        trades_by_timestamp[trade.timestamp][trade.symbol].append(trade)

    tradable_products = sorted(set(row.product for row in data.prices))

    listings = {product: {
        "symbol": product,
        "product": product,
        "denomination": 1,
    } for product in tradable_products}

    own_positions = defaultdict(int)
    own_trades = defaultdict(list)
    market_trades = defaultdict(list)

    profit_loss_by_product = defaultdict(float)

    for timestamp in sorted(prices_by_timestamp.keys()):
        order_depths = {}
        for product in tradable_products:
            row = prices_by_timestamp[timestamp][product]
            order_depths[product] = OrderDepth()

            for price, volume in zip(row.bid_prices, row.bid_volumes):
                order_depths[product].buy_orders[price] = volume

            for price, volume in zip(row.ask_prices, row.ask_volumes):
                order_depths[product].sell_orders[price] = -volume

        position = {product: position for product, position in own_positions.items() if position != 0}
        observations = Observation({}, {})

        state = TradingState(
            trader_data,
            timestamp,
            listings,
            order_depths,
            dict(own_trades),
            dict(market_trades),
            position,
            observations,
        )

        stdout = StringIO()
        stdout.close = lambda: None

        if print_output:
            with closing(Tee(stdout)):
                orders_by_symbol, conversions, trader_data = trader.run(state)
        else:
            with redirect_stdout(stdout):
                orders_by_symbol, conversions, trader_data = trader.run(state)

        for product in tradable_products:
            price = prices_by_timestamp[timestamp][product]

            profit_loss = profit_loss_by_product[product]
            if own_positions[product] != 0:
                profit_loss += own_positions[product] * price.mid_price

            result.activity_logs.append(create_activity_log_row(data.day, timestamp, product, price, profit_loss))

        sandbox_log = check_limits(tradable_products, orders_by_symbol, own_positions)

        result.sandbox_logs.append({
            "sandboxLog": sandbox_log,
            "lambdaLog": stdout.getvalue().rstrip(),
            "timestamp": timestamp,
        })

        current_market_trades = {}
        for product, trades in trades_by_timestamp[timestamp].items():
            current_market_trades[product] = [MarketTrade(t, t.quantity, t.quantity) for t in trades]

        for product in tradable_products:
            new_trades = []

            for order in orders_by_symbol.get(product, []):
                new_trades.extend(process_order(
                    timestamp,
                    order,
                    order_depths,
                    own_positions,
                    profit_loss_by_product,
                    current_market_trades.get(product, []),
                ))

            if len(new_trades) > 0:
                own_trades[product] = new_trades

        for product, trades in current_market_trades.items():
            for trade in trades:
                trade.trade.quantity = min(trade.buy_quantity, trade.sell_quantity)

            market_trades[product] = [t.trade for t in trades if t.trade.quantity > 0]

        current_trades = []
        for product in tradable_products:
            current_trades.extend([trade_to_dict(trade) for trade in own_trades[product]])
        for product in tradable_products:
            current_trades.extend([trade_to_dict(trade) for trade in trades_by_timestamp[timestamp][product]])

        result.trades.extend(current_trades)

    print_backtest_summary(result, tradable_products)
    return result
